from .http import HTTPCycle
from .websockets import WebSocketCycle
from .lifespan import LifespanCycleState, LifespanCycle

__all__ = [
    "HTTPCycle",
    "WebSocketCycle",
    "LifespanCycleState",
    "LifespanCycle",
]



import logging
from contextlib import ExitStack
from typing import (
    Any,
    ContextManager,
    Dict,
    Optional,
    TYPE_CHECKING,
)

from .exceptions import ConfigurationError
from .handlers import AbstractHandler
from .protocols import HTTPCycle, WebSocketCycle, LifespanCycle
from .backends import WebSocket
from .types import ASGIApp, WsRequest


if TYPE_CHECKING:  # pragma: no cover
    from awslambdaric.lambda_context import LambdaContext


DEFAULT_TEXT_MIME_TYPES = [
    "text/",
    "application/json",
    "application/javascript",
    "application/xml",
    "application/vnd.api+json",
]


logger = logging.getLogger("mangum")


class Mangum:
    """
    Creates an adapter instance.

    * **app** - An asynchronous callable that conforms to version 3.0 of the ASGI
    specification. This will usually be an ASGI framework application instance.
    * **lifespan** - A string to configure lifespan support. Choices are `auto`, `on`,
    and `off`. Default is `auto`.
    * **api_gateway_base_path** - Base path to strip from URL when using a custom
    domain name.
    * **text_mime_types** - A list of MIME types to include with the defaults that
    should not return a binary response in API Gateway.
    * **dsn** - A connection string required to configure a supported WebSocket backend.
    * **api_gateway_endpoint_url** - A string endpoint url to use for API Gateway when
    sending data to WebSocket connections. Default is to determine this automatically.
    * **api_gateway_region_name** - A string region name to use for API Gateway when
    sending data to WebSocket connections. Default is `AWS_REGION` environment variable.
    """

    app: ASGIApp
    lifespan: str = "auto"
    dsn: Optional[str] = None
    api_gateway_endpoint_url: Optional[str] = None
    api_gateway_region_name: Optional[str] = None

    def __init__(
        self,
        app: ASGIApp,
        lifespan: str = "auto",
        dsn: Optional[str] = None,
        api_gateway_endpoint_url: Optional[str] = None,
        api_gateway_region_name: Optional[str] = None,
        **handler_kwargs: Dict[str, Any]
    ) -> None:
        self.app = app
        self.lifespan = lifespan
        self.dsn = dsn
        self.api_gateway_endpoint_url = api_gateway_endpoint_url
        self.api_gateway_region_name = api_gateway_region_name
        self.handler_kwargs = handler_kwargs

        if self.lifespan not in ("auto", "on", "off"):
            raise ConfigurationError(
                "Invalid argument supplied for `lifespan`. Choices are: auto|on|off"
            )

    def __call__(self, event: dict, context: "LambdaContext") -> dict:
        logger.debug("Event received.")

        with ExitStack() as stack:
            if self.lifespan != "off":
                lifespan_cycle: ContextManager = LifespanCycle(self.app, self.lifespan)
                stack.enter_context(lifespan_cycle)

            handler = AbstractHandler.from_trigger(
                event, context, **self.handler_kwargs
            )
            request = handler.request

            if isinstance(request, WsRequest):
                api_gateway_endpoint_url = (
                    self.api_gateway_endpoint_url or handler.api_gateway_endpoint_url
                )
                websocket = WebSocket(
                    dsn=self.dsn,
                    api_gateway_endpoint_url=api_gateway_endpoint_url,
                    api_gateway_region_name=self.api_gateway_region_name,
                )
                websocket_cycle = WebSocketCycle(
                    request, handler.message_type, handler.connection_id, websocket
                )
                response = websocket_cycle(self.app, handler.body)
            else:
                http_cycle = HTTPCycle(request)
                response = http_cycle(self.app, handler.body)

        return handler.transform_response(response)



import base64
import urllib.parse
from typing import Dict, Any, TYPE_CHECKING

from .abstract_handler import AbstractHandler
from .. import Response, Request


if TYPE_CHECKING:  # pragma: no cover
    from awslambdaric.lambda_context import LambdaContext


class AwsApiGateway(AbstractHandler):
    """
    Handles AWS API Gateway events, transforming them into ASGI Scope and handling
    responses

    See: https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway.html
    """

    TYPE = "AWS_API_GATEWAY"

    def __init__(
        self,
        trigger_event: Dict[str, Any],
        trigger_context: "LambdaContext",
        api_gateway_base_path: str = "/",
        **kwargs: Dict[str, Any],  # type: ignore
    ):
        super().__init__(trigger_event, trigger_context, **kwargs)
        self.api_gateway_base_path = api_gateway_base_path

    @property
    def request(self) -> Request:
        event = self.trigger_event

        # multiValue versions of headers take precedence over their plain versions
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html#api-gateway-simple-proxy-for-lambda-input-format
        if event.get("multiValueHeaders"):
            headers = {
                k.lower(): ", ".join(v) if isinstance(v, list) else ""
                for k, v in event.get("multiValueHeaders", {}).items()
            }
        elif event.get("headers"):
            headers = {k.lower(): v for k, v in event.get("headers", {}).items()}
        else:
            headers = {}

        request_context = event["requestContext"]

        source_ip = request_context.get("identity", {}).get("sourceIp")

        path = event["path"]
        http_method = event["httpMethod"]

        if event.get("multiValueQueryStringParameters"):
            query_string = urllib.parse.urlencode(
                event.get("multiValueQueryStringParameters", {}), doseq=True
            ).encode()
        elif event.get("queryStringParameters"):
            query_string = urllib.parse.urlencode(
                event.get("queryStringParameters", {})
            ).encode()
        else:
            query_string = b""

        server_name = headers.get("host", "mangum")
        if ":" not in server_name:
            server_port = headers.get("x-forwarded-port", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))
        client = (source_ip, 0)

        if not path:
            path = "/"
        else:
            path = self._strip_base_path(path)

        return Request(
            method=http_method,
            headers=[[k.encode(), v.encode()] for k, v in headers.items()],
            path=urllib.parse.unquote(path),
            scheme=headers.get("x-forwarded-proto", "https"),
            query_string=query_string,
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    def _strip_base_path(self, path: str) -> str:
        if self.api_gateway_base_path and self.api_gateway_base_path != "/":
            if not self.api_gateway_base_path.startswith("/"):
                self.api_gateway_base_path = f"/{self.api_gateway_base_path}"
            if path.startswith(self.api_gateway_base_path):
                path = path[len(self.api_gateway_base_path) :]
        return path

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        headers, multi_value_headers = self._handle_multi_value_headers(
            response.headers
        )

        body, is_base64_encoded = self._handle_base64_response_body(
            response.body, headers
        )

        return {
            "statusCode": response.status,
            "headers": headers,
            "multiValueHeaders": multi_value_headers,
            "body": body,
            "isBase64Encoded": is_base64_encoded,
        }



import base64
import urllib.parse
from typing import Dict, Any

from . import AwsApiGateway
from .. import Response, Request


class AwsHttpGateway(AwsApiGateway):
    """
    Handles AWS HTTP Gateway events (v1.0 and v2.0), transforming them into ASGI Scope
    and handling responses

    See: https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html  # noqa: E501
    """

    TYPE = "AWS_HTTP_GATEWAY"

    @property
    def event_version(self) -> str:
        return self.trigger_event.get("version", "")

    @property
    def request(self) -> Request:
        event = self.trigger_event

        headers = {}
        if event.get("headers"):
            headers = {k.lower(): v for k, v in event.get("headers", {}).items()}

        request_context = event["requestContext"]

        # API Gateway v2
        if self.event_version == "2.0":
            source_ip = request_context["http"]["sourceIp"]
            path = request_context["http"]["path"]
            http_method = request_context["http"]["method"]
            query_string = event.get("rawQueryString", "").encode()

            if event.get("cookies"):
                headers["cookie"] = "; ".join(event.get("cookies", []))

        # API Gateway v1
        elif self.event_version == "1.0":
            # v1.0 of the HTTP Gateway supports multiValueHeaders
            if event.get("multiValueHeaders"):
                headers.update(
                    {
                        k.lower(): ", ".join(v) if isinstance(v, list) else ""
                        for k, v in event.get("multiValueHeaders", {}).items()
                    }
                )

            source_ip = request_context.get("identity", {}).get("sourceIp")

            path = event["path"]
            http_method = event["httpMethod"]

            # AWS Blog Post on this:
            # https://aws.amazon.com/blogs/compute/support-for-multi-value-parameters-in-amazon-api-gateway/  # noqa: E501
            # A multi value param will be in multi value _and_ regular
            # queryStringParameters. Multi value takes precedence.
            if event.get("multiValueQueryStringParameters", False):
                query_string = urllib.parse.urlencode(
                    event.get("multiValueQueryStringParameters", {}), doseq=True
                ).encode()
            elif event.get("queryStringParameters", False):
                query_string = urllib.parse.urlencode(
                    event.get("queryStringParameters", {})
                ).encode()
            else:
                query_string = b""
        else:
            raise RuntimeError(
                "Unsupported version of HTTP Gateway Spec, only v1.0 and v2.0 are "
                "supported."
            )

        server_name = headers.get("host", "mangum")
        if ":" not in server_name:
            server_port = headers.get("x-forwarded-port", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))
        client = (source_ip, 0)

        if not path:
            path = "/"
        else:
            path = self._strip_base_path(path)

        return Request(
            method=http_method,
            headers=[[k.encode(), v.encode()] for k, v in headers.items()],
            path=urllib.parse.unquote(path),
            scheme=headers.get("x-forwarded-proto", "https"),
            query_string=query_string,
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        """
        This handles some unnecessary magic from AWS

        >  API Gateway can infer the response format for you
        Boooooo

        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html#http-api-develop-integrations-lambda.response
        """
        headers, multi_value_headers = self._handle_multi_value_headers(
            response.headers
        )

        if self.event_version == "1.0":
            body, is_base64_encoded = self._handle_base64_response_body(
                response.body, headers
            )
            return {
                "statusCode": response.status,
                "headers": headers,
                "multiValueHeaders": multi_value_headers,
                "body": body,
                "isBase64Encoded": is_base64_encoded,
            }
        elif self.event_version == "2.0":
            # The API Gateway will infer stuff for us, but we'll just do that inference
            # here and keep the output consistent
            if "content-type" not in headers and response.body is not None:
                headers["content-type"] = "application/json"

            body, is_base64_encoded = self._handle_base64_response_body(
                response.body, headers
            )
            return {
                "statusCode": response.status,
                "headers": headers,
                "multiValueHeaders": multi_value_headers,
                "body": body,
                "isBase64Encoded": is_base64_encoded,
            }
        raise RuntimeError(  # pragma: no cover
            "Misconfigured event unable to return value, unsupported version."
        )



import base64
from abc import ABCMeta, abstractmethod
from typing import Dict, Any, TYPE_CHECKING, Tuple, List, Union

from ..types import Response, Request, WsRequest

if TYPE_CHECKING:  # pragma: no cover
    from awslambdaric.lambda_context import LambdaContext


class AbstractHandler(metaclass=ABCMeta):
    def __init__(
        self,
        trigger_event: Dict[str, Any],
        trigger_context: "LambdaContext",
        **kwargs: Dict[str, Any],
    ):
        self.trigger_event = trigger_event
        self.trigger_context = trigger_context

    @property
    @abstractmethod
    def request(self) -> Union[Request, WsRequest]:
        """
        Parse an ASGI scope from the request event
        """

    @property
    @abstractmethod
    def body(self) -> bytes:
        """
        Get the raw body from the request event
        """

    @abstractmethod
    def transform_response(self, response: Response) -> Dict[str, Any]:
        """
        After running our application, transform the response to the correct format for
        this handler
        """

    @property
    def message_type(self) -> str:
        request_context = self.trigger_event["requestContext"]
        return request_context["eventType"]

    @property
    def connection_id(self) -> str:
        request_context = self.trigger_event["requestContext"]
        return request_context["connectionId"]

    @property
    def api_gateway_endpoint_url(self) -> str:
        request_context = self.trigger_event["requestContext"]
        domain = request_context["domainName"]
        stage = request_context["stage"]
        api_gateway_endpoint_url = f"https://{domain}/{stage}/@connections"

        return api_gateway_endpoint_url

    @staticmethod
    def from_trigger(
        trigger_event: Dict[str, Any],
        trigger_context: "LambdaContext",
        **kwargs: Dict[str, Any],
    ) -> "AbstractHandler":
        """
        A factory method that determines which handler to use. All this code should
        probably stay in one place to make sure we are able to uniquely find each
        handler correctly.
        """

        # These should be ordered from most specific to least for best accuracy
        if (
            "requestContext" in trigger_event
            and "elb" in trigger_event["requestContext"]
        ):
            from . import AwsAlb

            return AwsAlb(trigger_event, trigger_context, **kwargs)

        if (
            "requestContext" in trigger_event
            and "connectionId" in trigger_event["requestContext"]
        ):
            from . import AwsWsGateway

            return AwsWsGateway(
                trigger_event, trigger_context, **kwargs  # type: ignore
            )

        if (
            "Records" in trigger_event
            and len(trigger_event["Records"]) > 0
            and "cf" in trigger_event["Records"][0]
        ):
            from . import AwsCfLambdaAtEdge

            return AwsCfLambdaAtEdge(trigger_event, trigger_context, **kwargs)

        if "version" in trigger_event and "requestContext" in trigger_event:
            from . import AwsHttpGateway

            return AwsHttpGateway(
                trigger_event, trigger_context, **kwargs  # type: ignore
            )

        if "resource" in trigger_event:
            from . import AwsApiGateway

            return AwsApiGateway(
                trigger_event, trigger_context, **kwargs  # type: ignore
            )

        raise TypeError("Unable to determine handler from trigger event")

    @staticmethod
    def _handle_multi_value_headers(
        response_headers: List[List[bytes]],
    ) -> Tuple[Dict[str, str], Dict[str, List[str]]]:
        headers: Dict[str, str] = {}
        multi_value_headers: Dict[str, List[str]] = {}
        for key, value in response_headers:
            lower_key = key.decode().lower()
            if lower_key in multi_value_headers:
                multi_value_headers[lower_key].append(value.decode())
            elif lower_key in headers:
                # Move existing to multi_value_headers and append current
                multi_value_headers[lower_key] = [
                    headers[lower_key],
                    value.decode(),
                ]
                del headers[lower_key]
            else:
                headers[lower_key] = value.decode()
        return headers, multi_value_headers

    @staticmethod
    def _handle_base64_response_body(
        body: bytes, headers: Dict[str, str]
    ) -> Tuple[str, bool]:
        """
        To ease debugging for our users, try and return strings where we can,
        otherwise to ensure maximum compatibility with binary data, base64 encode it.
        """
        is_base64_encoded = False
        output_body = ""
        if body != b"":
            from ..adapter import DEFAULT_TEXT_MIME_TYPES

            for text_mime_type in DEFAULT_TEXT_MIME_TYPES:
                if text_mime_type in headers.get("content-type", ""):
                    try:
                        output_body = body.decode()
                    except UnicodeDecodeError:
                        # Can't decode it, base64 it and be done
                        output_body = base64.b64encode(body).decode()
                        is_base64_encoded = True
                    break
            else:
                # Not text, base64 encode
                output_body = base64.b64encode(body).decode()
                is_base64_encoded = True

        return output_body, is_base64_encoded



import enum
import asyncio
import logging
from io import BytesIO
from dataclasses import dataclass

from ..backends import WebSocket
from ..exceptions import UnexpectedMessage, WebSocketClosed, WebSocketError
from ..types import ASGIApp, Message, WsRequest, Response


class WebSocketCycleState(enum.Enum):
    """
    The state of the ASGI WebSocket connection.

    * **CONNECTING** - Initial state. The ASGI application instance will be run with the
    connection scope containing the `websocket` type.
    * **HANDSHAKE** - The ASGI `websocket` connection with the application has been
    established, and a `websocket.connect` event has been pushed to the application
    queue. The application will respond by accepting or rejecting the connection.
    If rejected, a 403 response will be returned to the client, and it will be removed
    from API Gateway.
    * **RESPONSE** - Handshake accepted by the application. Data received in the API
    Gateway message event will be sent to the application. A `websocket.receive` event
    will be pushed to the application queue.
    * **DISCONNECTING** - The ASGI connection cycle is complete and should be
    disconnected from the application. A `websocket.disconnect` event will be pushed to
    the queue, and a response will be returned to the client connection.
    * **CLOSED** - The application has sent a `websocket.close` message. This will
    either be in response to a `websocket.disconnect` event or occurs when a connection
    is rejected in response to a `websocket.connect` event.
    """

    CONNECTING = enum.auto()
    HANDSHAKE = enum.auto()
    RESPONSE = enum.auto()
    DISCONNECTING = enum.auto()
    CLOSED = enum.auto()


@dataclass
class WebSocketCycle:
    """
    Manages the application cycle for an ASGI `websocket` connection.

    * **websocket** - A `WebSocket` connection handler interface for the selected
    `WebSocketBackend` subclass. Contains the ASGI connection `scope` and client
    connection identifier.
    * **state** - An enumerated `WebSocketCycleState` type that indicates the state of
    the ASGI connection.
    * **app_queue** - An asyncio queue (FIFO) containing messages to be received by the
    application.
    """

    request: WsRequest
    message_type: str
    connection_id: str
    websocket: WebSocket
    state: WebSocketCycleState = WebSocketCycleState.CONNECTING

    def __post_init__(self) -> None:
        self.logger: logging.Logger = logging.getLogger("mangum.websocket")
        self.loop = asyncio.get_event_loop()
        self.app_queue: asyncio.Queue = asyncio.Queue()
        self.body: BytesIO = BytesIO()
        self.response: Response = Response(200, [], b"")

    def __call__(self, app: ASGIApp, initial_body: bytes) -> Response:
        self.logger.debug("WebSocket cycle starting.")
        self.initial_body = initial_body

        if self.message_type == "CONNECT":
            scope = self.request.scope
            del scope["aws.event"]
            del scope["aws.context"]
            self.loop.run_until_complete(
                self.websocket.on_connect(self.connection_id, scope)
            )
        elif self.message_type == "MESSAGE":
            self.app_queue.put_nowait({"type": "websocket.connect"})
            asgi_instance = self.run(app)
            asgi_task = self.loop.create_task(asgi_instance)
            self.loop.run_until_complete(asgi_task)
        elif self.message_type == "DISCONNECT":
            self.loop.run_until_complete(
                self.websocket.on_disconnect(self.connection_id)
            )

        return self.response

    async def run(self, app: ASGIApp) -> None:
        """
        Calls the application with the `websocket` connection scope.
        """
        self.scope = await self.websocket.on_message(self.connection_id)
        scope = self.scope.copy()  # type: ignore
        scope.update(
            {
                "aws.event": self.request.trigger_event,
                "aws.context": self.request.trigger_context,
            }
        )
        try:
            await app(scope, self.receive, self.send)
        except WebSocketClosed:
            self.response.status = 403
        except UnexpectedMessage:
            self.response.status = 500
        except BaseException as exc:
            self.logger.error("Exception in ASGI application", exc_info=exc)
            self.response.status = 500

    async def receive(self) -> Message:
        """
        Awaited by the application to receive ASGI `websocket` events.
        """
        if self.state is WebSocketCycleState.CONNECTING:

            # Initial ASGI connection established. The next event returned by the queue
            # will be `websocket.connect` to initiate the handshake.
            self.state = WebSocketCycleState.HANDSHAKE

        elif self.state is WebSocketCycleState.HANDSHAKE:

            # ASGI connection handshake accepted. The next event returned by the queue
            # will be `websocket.receive` containing the message data from API Gateway.
            self.state = WebSocketCycleState.RESPONSE

        return await self.app_queue.get()

    async def send(self, message: Message) -> None:
        """
        Awaited by the application to send ASGI `websocket` events.
        """
        message_type = message["type"]
        self.logger.info(
            "%s:  '%s' event received from application.", self.state, message_type
        )

        if self.state is WebSocketCycleState.HANDSHAKE and message_type in (
            "websocket.accept",
            "websocket.close",
        ):

            # API Gateway handles the WebSocket client handshake in the connect event,
            # and it cannot be negotiated by the application directly. The application
            # may choose to close the connection at this point. This process does not
            # support subprotocols.
            if message_type == "websocket.accept":
                await self.app_queue.put(
                    {
                        "type": "websocket.receive",
                        "bytes": None,
                        "text": self.initial_body.decode(),
                    }
                )
            elif message_type == "websocket.close":
                self.state = WebSocketCycleState.CLOSED
                await self.websocket.delete_connection(self.connection_id)
                raise WebSocketClosed

        elif (
            self.state is WebSocketCycleState.RESPONSE
            and message_type == "websocket.close"
        ):

            # The application is explicitly closing the connection. It should be
            # disconnected and removed in API Gateway.
            await self.websocket.delete_connection(self.connection_id)

        elif (
            self.state is WebSocketCycleState.RESPONSE
            and message_type == "websocket.send"
        ):

            # The application requested to send some data in response to the
            # "websocket.receive" event. After this, a "websocket.disconnect"
            # event is pushed to let the application finish gracefully.
            # Then the lambda's execution is ended.

            if message.get("body") is not None:
                raise WebSocketError(
                    "Application attemped to send a binary payload, "
                    "but it's unsupported!"
                )

            message_text = message.get("text", "")
            body = message_text.encode()

            await self.websocket.post_to_connection(self.connection_id, body=body)
            await self.app_queue.put({"type": "websocket.disconnect", "code": 1000})

        else:
            raise UnexpectedMessage(
                f"{self.state}: Unexpected '{message_type}' event received."
            )



from dataclasses import dataclass, field
from typing import (
    List,
    Tuple,
    Dict,
    Any,
    Union,
    Optional,
    MutableMapping,
    Awaitable,
    Callable,
    TYPE_CHECKING,
)
from typing_extensions import Protocol


Message = MutableMapping[str, Any]
Scope = MutableMapping[str, Any]
Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]


if TYPE_CHECKING:  # pragma: no cover
    from awslambdaric.lambda_context import LambdaContext


class ASGIApp(Protocol):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...  # pragma: no cover


@dataclass
class BaseRequest:
    """
    A holder for an ASGI scope. Contains additional meta from the event that triggered
    the Lambda function.
    """

    headers: List[List[bytes]]
    path: str
    scheme: str
    query_string: bytes
    server: Tuple[str, int]
    client: Tuple[str, int]

    # Invocation event
    trigger_event: Dict[str, Any]
    trigger_context: Union["LambdaContext", Dict[str, Any]]
    event_type: str

    http_version: str = "1.1"
    raw_path: Optional[str] = None
    root_path: str = ""
    asgi: Dict[str, str] = field(default_factory=lambda: {"version": "3.0"})

    def scope(self) -> Scope:
        return {
            "http_version": self.http_version,
            "headers": self.headers,
            "path": self.path,
            "raw_path": self.raw_path,
            "root_path": self.root_path,
            "scheme": self.scheme,
            "query_string": self.query_string,
            "server": self.server,
            "client": self.client,
            "asgi": self.asgi,
            # Meta data to pass along to the application in case they need it
            "aws.event": self.trigger_event,
            "aws.context": self.trigger_context,
            "aws.eventType": self.event_type,
        }


@dataclass
class Request(BaseRequest):
    """
    A holder for an ASGI scope. Specific for usage with HTTP connections.

    https://asgi.readthedocs.io/en/latest/specs/www.html#http-connection-scope
    """

    type: str = "http"
    method: str = "GET"

    @property
    def scope(self) -> Scope:
        scope = super().scope()
        scope.update({"type": self.type, "method": self.method})
        return scope


@dataclass
class WsRequest(BaseRequest):
    """
    A holder for an ASGI scope. Specific for usage with WebSocket connections.

    https://asgi.readthedocs.io/en/latest/specs/www.html#websocket-connection-scope
    """

    type: str = "websocket"
    subprotocols: List[str] = field(default_factory=lambda: [])

    @property
    def scope(self) -> Scope:
        scope = super().scope()
        scope.update({"type": self.type, "subprotocols": self.subprotocols})
        return scope


@dataclass
class Response:
    status: int
    headers: List[List[bytes]]  # ex: [[b'content-type', b'text/plain; charset=utf-8']]
    body: bytes



from dataclasses import dataclass
from typing import Any


@dataclass
class WebSocketBackend:  # pragma: no cover
    """
    Base class for implementing WebSocket backends to store API Gateway connections.

    Data source backends are required to implement configuration using the `dsn`
    connection string setting.
    """

    dsn: str

    async def __aenter__(self) -> "WebSocketBackend":
        """
        Establish the connection to a data source.
        """
        raise NotImplementedError()

    async def __aexit__(self, *exc_info: Any) -> None:
        """
        Closes the connection to a data source.
        """
        raise NotImplementedError()

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        """
        Save the JSON scope for a connection.
        """
        raise NotImplementedError()

    async def retrieve(self, connection_id: str) -> str:
        """
        Retrieve the JSON scope for a connection.
        """
        raise NotImplementedError()

    async def delete(self, connection_id: str) -> None:
        """
        Delete the JSON scope for a connection.
        """
        raise NotImplementedError()



import asyncio
import logging
from typing import Dict, Optional
import json
from functools import partial
from dataclasses import dataclass, InitVar
from urllib.parse import urlparse

try:
    import httpx
except ImportError:  # pragma: no cover
    httpx = None  # type: ignore

try:
    import boto3
    from botocore.auth import SigV4Auth
    from botocore.awsrequest import AWSRequest
except ImportError:  # pragma: no cover
    boto3 = None  # type: ignore


from .base import WebSocketBackend
from ..exceptions import WebSocketError, ConfigurationError
from ..types import Scope


def get_sigv4_headers(
    method: str,
    url: str,
    data: Optional[bytes] = None,
    region_name: Optional[str] = None,
) -> Dict:
    session = boto3.Session()
    credentials = session.get_credentials()
    creds = credentials.get_frozen_credentials()
    region = region_name or session.region_name

    request = AWSRequest(method=method, url=url, data=data)
    SigV4Auth(creds, "execute-api", region).add_auth(request)

    return dict(request.headers)


@dataclass
class WebSocket:
    """
    A `WebSocket` connection handler interface for the
    selected `WebSocketBackend` subclass
    """

    dsn: InitVar[Optional[str]]
    api_gateway_endpoint_url: str
    api_gateway_region_name: Optional[str] = None

    def __post_init__(self, dsn: Optional[str]) -> None:
        if boto3 is None:  # pragma: no cover
            raise WebSocketError("boto3 must be installed to use WebSockets.")

        if httpx is None:  # pragma: no cover
            raise WebSocketError("httpx must be installed to use WebSockets.")

        if dsn is None:
            raise ConfigurationError(
                "The `dsn` parameter must be provided for WebSocket connections."
            )

        self.logger: logging.Logger = logging.getLogger("mangum.backends")
        parsed_dsn = urlparse(dsn)
        if not any((parsed_dsn.hostname, parsed_dsn.path)):
            raise ConfigurationError("Invalid value for `dsn` provided.")

        scheme = parsed_dsn.scheme
        self.logger.debug(
            f"Attempting WebSocket backend connection using scheme: {scheme}"
        )

        if scheme == "sqlite":
            self.logger.info(
                "The `SQLiteBackend` should be only be used for local "
                "debugging. It will not work in a deployed environment."
            )
            from mangum.backends.sqlite import SQLiteBackend

            self._Backend = SQLiteBackend  # type: ignore

        elif scheme == "dynamodb":
            from mangum.backends.dynamodb import DynamoDBBackend

            self._Backend = DynamoDBBackend  # type: ignore

        elif scheme == "s3":
            from mangum.backends.s3 import S3Backend

            self._Backend = S3Backend  # type: ignore

        elif scheme in ("postgresql", "postgres"):
            from mangum.backends.postgresql import PostgreSQLBackend

            self._Backend = PostgreSQLBackend  # type: ignore

        elif scheme == "redis":
            from mangum.backends.redis import RedisBackend

            self._Backend = RedisBackend  # type: ignore

        else:
            raise ConfigurationError(f"{scheme} does not match a supported backend.")

        self.dsn = dsn

        self.logger.info("WebSocket backend connection established.")

    async def load_scope(self, backend: WebSocketBackend, connection_id: str) -> Scope:
        loaded_scope = await backend.retrieve(connection_id)
        scope = json.loads(loaded_scope)
        scope.update(
            {
                "query_string": scope["query_string"].encode(),
                "headers": [[h[0].encode(), h[1].encode()] for h in scope["headers"]],
                "client": tuple(scope["client"]),
                "server": tuple(scope["server"]),
            }
        )

        return scope

    async def save_scope(
        self, backend: WebSocketBackend, connection_id: str, scope: Scope
    ) -> None:
        scope.update(
            {
                "query_string": scope["query_string"].decode(),
                "headers": [[h[0].decode(), h[1].decode()] for h in scope["headers"]],
            }
        )
        json_scope = json.dumps(scope)
        await backend.save(connection_id, json_scope=json_scope)

    async def on_connect(self, connection_id: str, initial_scope: Scope) -> None:
        self.logger.debug("Creating scope entry for %s", connection_id)
        async with self._Backend(self.dsn) as backend:  # type: ignore
            await self.save_scope(backend, connection_id, initial_scope)

    async def on_message(self, connection_id: str) -> Scope:
        self.logger.debug("Retrieving scope entry for %s", connection_id)
        async with self._Backend(self.dsn) as backend:  # type: ignore
            scope = await self.load_scope(backend, connection_id)
        return scope

    async def on_disconnect(self, connection_id: str) -> None:
        self.logger.debug("Deleting scope entry for %s", connection_id)
        async with self._Backend(self.dsn) as backend:  # type: ignore
            await backend.delete(connection_id)

    async def post_to_connection(self, connection_id: str, body: bytes) -> None:
        async with httpx.AsyncClient() as client:
            await self._post_to_connection(connection_id, client=client, body=body)

    async def delete_connection(self, connection_id: str) -> None:
        async with httpx.AsyncClient() as client:
            await self._request_to_connection("DELETE", connection_id, client=client)

    async def _post_to_connection(
        self,
        connection_id: str,
        *,
        client: "httpx.AsyncClient",
        body: bytes,
    ) -> None:  # pragma: no cover
        response = await self._request_to_connection(
            "POST", connection_id, client=client, body=body
        )

        if response.status_code == 410:
            await self.on_disconnect(connection_id)
        elif response.status_code != 200:
            raise WebSocketError(f"Error: {response.status_code}")

    async def _request_to_connection(
        self,
        method: str,
        connection_id: str,
        *,
        client: "httpx.AsyncClient",
        body: Optional[bytes] = None,
    ) -> "httpx.Response":
        loop = asyncio.get_event_loop()
        url = f"{self.api_gateway_endpoint_url}/{connection_id}"
        headers = await loop.run_in_executor(
            None,
            partial(
                get_sigv4_headers,
                method,
                url,
                body,
                self.api_gateway_region_name,
            ),
        )

        return await client.request(method, url, content=body, headers=headers)



import base64
from typing import Dict, Any, List

from .abstract_handler import AbstractHandler
from .. import Response, Request


class AwsCfLambdaAtEdge(AbstractHandler):
    """
    Handles AWS Elastic Load Balancer, really Application Load Balancer events
    transforming them into ASGI Scope and handling responses

    See: https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/lambda-event-structure.html  # noqa: E501
    """

    TYPE = "AWS_CF_LAMBDA_AT_EDGE"

    @property
    def request(self) -> Request:
        event = self.trigger_event

        cf_request = event["Records"][0]["cf"]["request"]

        scheme_header = cf_request["headers"].get("cloudfront-forwarded-proto", [{}])
        scheme = scheme_header[0].get("value", "https")

        host_header = cf_request["headers"].get("host", [{}])
        server_name = host_header[0].get("value", "mangum")
        if ":" not in server_name:
            forwarded_port_header = cf_request["headers"].get("x-forwarded-port", [{}])
            server_port = forwarded_port_header[0].get("value", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))

        source_ip = cf_request["clientIp"]
        client = (source_ip, 0)

        return Request(
            method=cf_request["method"],
            headers=[
                [k.encode(), v[0]["value"].encode()]
                for k, v in cf_request["headers"].items()
            ],
            path=cf_request["uri"],
            scheme=scheme,
            query_string=cf_request["querystring"].encode(),
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    @property
    def body(self) -> bytes:
        request = self.trigger_event["Records"][0]["cf"]["request"]
        body = request.get("body", {}).get("data", None) or b""

        if request.get("body", {}).get("encoding", "") == "base64":
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        headers_dict, _ = self._handle_multi_value_headers(response.headers)
        body, is_base64_encoded = self._handle_base64_response_body(
            response.body, headers_dict
        )

        # Expand headers to weird list of Dict[str, List[Dict[str, str]]]
        headers_expanded: Dict[str, List[Dict[str, str]]] = {
            key.decode().lower(): [{"key": key.decode().lower(), "value": val.decode()}]
            for key, val in response.headers
        }
        return {
            "status": response.status,
            "headers": headers_expanded,
            "body": body,
            "isBase64Encoded": is_base64_encoded,
        }



from typing import Any, Dict, Tuple
import base64


from ..types import Response, WsRequest
from .abstract_handler import AbstractHandler


def get_server_and_headers(event: dict) -> Tuple:  # pragma: no cover
    if event.get("multiValueHeaders"):
        headers = {
            k.lower(): ", ".join(v) if isinstance(v, list) else ""
            for k, v in event.get("multiValueHeaders", {}).items()
        }
    elif event.get("headers"):
        headers = {k.lower(): v for k, v in event.get("headers", {}).items()}
    else:
        headers = {}

    # Subprotocols are not supported, so drop Sec-WebSocket-Protocol to be safe
    headers.pop("sec-websocket-protocol", None)

    server_name = headers.get("host", "mangum")
    if ":" not in server_name:
        server_port = headers.get("x-forwarded-port", 80)
    else:
        server_name, server_port = server_name.split(":")
    server = (server_name, int(server_port))

    return server, headers


class AwsWsGateway(AbstractHandler):
    """
    Handles AWS API Gateway Websocket events, transforming
    them into ASGI Scope and handling responses

    See: https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html#api-gateway-simple-proxy-for-lambda-input-format  # noqa: E501
    """

    TYPE = "AWS_WS_GATEWAY"

    @property
    def request(self) -> WsRequest:
        request_context = self.trigger_event["requestContext"]
        server, headers = get_server_and_headers(self.trigger_event)
        source_ip = request_context.get("identity", {}).get("sourceIp")
        client = (source_ip, 0)
        headers_list = [[k.encode(), v.encode()] for k, v in headers.items()]

        return WsRequest(
            headers=headers_list,
            path="/",
            scheme=headers.get("x-forwarded-proto", "wss"),
            query_string=b"",
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        return {"statusCode": response.status}
